from typing import Callable, Set
from px_settings.contrib.django import settings as s
from dataclasses import dataclass, field
from functools import lru_cache
from django.utils.module_loading import import_string

from .const import LogOperate


__all__ = 'NAME', 'Settings', 'settings',

NAME = 'PXD_ACTIONS_TRACKER'
cached_import = lru_cache()(import_string)


@s(NAME)
@dataclass
class Settings:
    OPERATE: LogOperate = LogOperate.STRAIGHT
    SAVER: str = 'pxd_actions_tracker.storages.db.storage.save'
    SENSITIVE_FIELDS: Set[str] = field(default_factory=lambda: {
        'api',
        'token',
        'key',
        'secret',
        'password',
        'signature',
        'authorization',
    })
    SENSITIVE_PLACEHOLDER: str = '*******'

    def get_saver(self) -> Callable:
        return cached_import(self.SAVER)


settings = Settings()



# Generated by Django 3.2.8 on 2021-12-27 14:53

import django.contrib.postgres.fields
from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Log',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, verbose_name='ID')),
                ('domain', models.CharField(max_length=1024, verbose_name='Domain')),
                ('tracking', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=128), blank=True, default=list, size=None, verbose_name='Tracking codes stack')),
                ('message', models.TextField(blank=True, default='', verbose_name='Message')),
                ('detail', models.JSONField(blank=True, default=dict, verbose_name='Detail')),
                ('created_at', models.DateTimeField(verbose_name='Created at')),
            ],
            options={
                'verbose_name': 'Log entry',
                'verbose_name_plural': 'Log entries',
                'ordering': ('-created_at', 'domain'),
            },
        ),
    ]



from typing import Callable, Dict, Optional, Sequence, Tuple
import json
import traceback
import logging
from uuid import UUID
from functools import wraps
from px_domains import Domain
from django.http import RawPostDataException

from django.http import HttpRequest, HttpResponse

from ..cases import log
from ..utils import clean_data, get_ip_address
from ..domains import ERROR, INFO, VIEW


__all__ = (
    'MESSAGE_FORMAT',

    'track',
    'track_middleware',
    'decorate',
    'create_middleware',

    'should_track_non_200',
    'should_track_all',
)


logger = logging.getLogger(__name__)


HEADER_PREFIX = 'HTTP_'
HEADER_PREFIX_LENGTH = len(HEADER_PREFIX)
MESSAGE_FORMAT = '[VISIT]: "{request[path]}"'.format


def resolve_request_data(request: HttpRequest) -> Dict:
    user = getattr(request, 'user', None)
    user = user if user and not user.is_anonymous else None

    data_drf = request.data if hasattr(request, 'data') else {}
    data_post = dict(request.POST.lists())
    data_body = {}

    try:
        data_body = (
            json.loads(request.body.decode('utf-8') or '{}')
            if (
                (
                    request.META.get('HTTP_ACCEPT') == 'application/json'
                    or
                    getattr(request, 'content_type', None) == 'application/json'
                    or
                    request.META.get('CONTENT_TYPE') == 'application/json'
                )
                and
                hasattr(request, 'body')
            ) else
            {}
        )
    except RawPostDataException as e:
        pass
    except Exception as e:
        logger.exception(e)

    return {
        'ip': get_ip_address(request),
        'host': request.get_host(),
        'method': request.method,
        'path': request.path,
        'content_type': request.content_type,
        'user': (
            {'id': None, 'name': 'Anonymous'}
            if user is None else
            {'id': user.id, 'name': user.get_username()}
        ),
        'data': clean_data({**data_drf, **data_post, **data_body}),
        'query_params': clean_data(dict(request.GET.lists())),
        'headers': clean_data({
            key[HEADER_PREFIX_LENGTH:]: value
            for key, value in request.META.items()
            if key.startswith(HEADER_PREFIX)
        })
    }


def resolve_response_data(response: HttpResponse) -> Dict:
    rendered_content = None

    if response.streaming:
        rendered_content = None
    elif hasattr(response, 'rendered_content'):
        rendered_content = response.rendered_content
    else:
        rendered_content = response.content

    return {
        'content': clean_data(rendered_content),
        'status_code': response.status_code,
    }


def resolve_exception_data(
    error: Optional[Exception],
    response: Optional[HttpResponse]
) -> Optional[Dict]:
    if error is not None:
        return {'message': str(error), 'traceback': traceback.format_exc()}
    if response is not None and response.status_code // 100 == 5:
        return {
            'message': 'Error',
            'traceback': (
                ''.join(traceback.format_list(response._init_stack))
                if hasattr(response, '_init_stack') else
                ''
            )
        }

    return None


def should_track_non_200(
    request: HttpRequest,
    response: Optional[HttpResponse] = None,
    error: Optional[Exception] = None,
    args: Sequence = (),
    kwargs: Dict = {},
) -> Tuple[bool, str]:
    should = (
        error is not None
        or
        response is None
        or
        response.status_code // 100 in (4, 5)
    )

    return should, ERROR


def should_track_all(
    request: HttpRequest,
    response: Optional[HttpResponse] = None,
    error: Optional[Exception] = None,
    args: Sequence = (),
    kwargs: Dict = {},
) -> Tuple[bool, str]:
    e, _ = should_track_non_200(request, response=response, error=error)

    return True, ERROR if e else INFO


def resolve_log_detail(
    request: HttpRequest,
    response: Optional[HttpResponse] = None,
    error: Optional[Exception] = None,
    args: Sequence = (),
    kwargs: Dict = {},
) -> Dict:
    request_data = resolve_request_data(request)
    response_data = response and resolve_response_data(response) or None
    error_data = resolve_exception_data(error, response)

    return {
        'request': request_data, 'response': response_data,
        'error': error_data,
    }


def track(
    domain: str,
    request: HttpRequest,
    response: Optional[HttpResponse] = None,
    error: Optional[Exception] = None,
    args: Sequence = (),
    kwargs: Dict = {},
    message_format: Callable = MESSAGE_FORMAT,
    should_track: Callable = should_track_non_200,
    detail_resolver: Callable = resolve_log_detail,
) -> Optional[UUID]:
    kw = dict(response=response, error=error, args=args, kwargs=kwargs)
    should, domain_postfix = should_track(request, **kw)

    if not should:
        return None

    try:
        detail = detail_resolver(request, **kw)

        return log(
            (domain | domain_postfix) if domain_postfix else domain,
            message_format(**detail), detail=detail,
        )
    except Exception as e:
        logger.exception(e)

    return None


def track_middleware(
    domain: str,
    get_response: Callable,
    request: HttpRequest,
    tracker: Callable = track,
    args: Sequence = [],
    kwargs: dict = {},
) -> Tuple[HttpResponse, Optional[UUID]]:
    error = None
    response = None
    log_id = None

    try:
        response = get_response(request, *args, **kwargs)
    except Exception as e:
        error = e

    try:
        log_id = tracker(
            domain,
            request,
            response=response,
            error=error,
            args=args,
            kwargs=kwargs,
        )
    except Exception as e:
        logger.exception(e)

    if error is not None:
        raise error

    return response, log_id


def decorate(domain: Domain = VIEW, tracker: Callable = track):
    def decorator(f: Callable):
        @wraps(f)
        def runner(request: HttpRequest, *args, **kwargs):
            response, _ = track_middleware(
                domain, f, request, tracker=tracker, args=args, kwargs=kwargs
            )

            return response

        return runner

    return decorator


def create_middleware(base_domain: Domain = VIEW, tracker: Callable = track):
    def middleware(get_response):
        def runner(request: HttpRequest, *args, **kwargs):
            domain = base_domain

            if request.resolver_match is not None:
                domain = domain | Domain(request.resolver_match.view_name)

            response, _ = track_middleware(
                domain, get_response, request, tracker=tracker,
                args=args, kwargs=kwargs
            )

            return response

        return runner

    return middleware



import logging
from typing import Optional
from copy import copy
from px_domains import Domain

from django.conf import settings
from django.utils.module_loading import import_string


__all__ = 'EXCEPTION_LOG_DOMAIN', 'ActionsTrackerHandler',

logger = logging.getLogger(__name__)
EXCEPTION_LOG_DOMAIN = Domain('LOG')


class ActionsTrackerHandler(logging.Handler):
    domain: Domain = EXCEPTION_LOG_DOMAIN

    def __init__(self, domain: Optional[Domain] = None, reporter_class=None):
        super().__init__()
        self.domain = domain if domain is not None else self.domain
        self.reporter_class = import_string(
            reporter_class
            or
            getattr(settings, 'DEFAULT_EXCEPTION_REPORTER', None)
            or
            'django.views.debug.ExceptionReporter'
        )

    def emit(self, record):
        try:
            request = record.request
        except Exception:
            request = None

        # Since we add a nicely formatted traceback on our own, create a copy
        # of the log record without the exception data.
        no_exc_record = copy(record)
        no_exc_record.exc_info = None
        no_exc_record.exc_text = None

        if record.exc_info:
            exc_info = record.exc_info
        else:
            exc_info = (None, record.getMessage(), None)

        reporter = self.reporter_class(request, is_email=False, *exc_info)
        message = self.format(no_exc_record)
        traceback = reporter.get_traceback_text()
        self.log(
            self.domain | record.levelname | record.name,
            '%s: %s' % (record.levelname, message),
            {
                'request': self.resolve_request_data(request) if request else None,
                'message': message,
                'traceback': traceback,
            }
        )

    def resolve_request_data(self, request):
        from . import track_views

        return track_views.resolve_request_data(request)

    def log(self, domain, message, detail):
        from ..cases import log

        return log(domain, message.format(**detail), detail=detail)



from django.db import models
from django.contrib.postgres.fields import ArrayField
from django.utils.translation import pgettext_lazy
import uuid


class Log(models.Model):
    class Meta:
        verbose_name = pgettext_lazy('pxd-actions-tracker', 'Log entry')
        verbose_name_plural = pgettext_lazy('pxd-actions-tracker', 'Log entries')
        ordering = '-created_at', 'domain'

    id = models.UUIDField(
        verbose_name=pgettext_lazy('pxd-actions-tracker', 'ID'),
        primary_key=True,
        default=uuid.uuid4,
        editable=False
    )

    domain = models.CharField(
        verbose_name=pgettext_lazy('pxd-actions-tracker', 'Domain'),
        max_length=1024, null=False
    )
    tracking = ArrayField(
        models.CharField(max_length=128),
        verbose_name=pgettext_lazy('pxd-actions-tracker', 'Tracking codes stack'),
        null=False, blank=True, default=list,
    )
    message = models.TextField(
        verbose_name=pgettext_lazy('pxd-actions-tracker', 'Message'),
        null=False, blank=True, default=''
    )
    detail = models.JSONField(
        verbose_name=pgettext_lazy('pxd-actions-tracker', 'Detail'),
        null=False, blank=True, default=dict
    )

    created_at = models.DateTimeField(
        verbose_name=pgettext_lazy('pxd-actions-tracker', 'Created at'),
        null=False
    )

    def __str__(self):
        return f'Log #{self.id}: {self.message}'



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy

__all__ = ('DBActionsTrackerConfig',)


class DBActionsTrackerConfig(AppConfig):
    name = 'pxd_actions_tracker.storages.db'
    label = 'pxd_actions_tracker_storages_db'
    verbose_name = pgettext_lazy('pxd-actions-tracker', 'Actions tracker: Database')



from .cases import uuid_track


__all__ = 'tracking_middleware',


def tracking_middleware(get_response):
    def runner(request):
        with uuid_track():
            response = get_response(request)
        return response

    return runner



from django.dispatch import Signal


log_created: Signal = Signal()
"""Fires when new log entry created.

Args:
    instance: Instance of created log object.
"""

tracking_enter: Signal = Signal()
"""Fires when new tracking code added to a stack.

Args:
    code (str): Tracking code that was added.
"""

tracking_exit: Signal = Signal()
"""Fires when new tracking code removed from a stack.

Args:
    code (str): Tracking code that was removed.
"""



from uuid import uuid4
from contextlib import ContextDecorator

from ..const import TRACKING
from ..signals import tracking_enter, tracking_exit


__all__ = 'track', 'uuid_track',


class track(ContextDecorator):
    def __init__(self, tracking_code: str):
        assert tracking_code != '', (
            'Tracking code must be anything useful, not just an empty string.'
        )

        self.tracking_code = tracking_code
        self.token = None

    def __enter__(self):
        current = TRACKING.get() or []

        self.token = TRACKING.set(current + [self.tracking_code])
        tracking_enter.send(self, code=self.tracking_code)

        return self

    def __exit__(self, *exc):
        TRACKING.reset(self.token)
        tracking_exit.send(self, code=self.tracking_code)

        return False


def uuid_track(*args, **kwargs):
    return track(str(uuid4()), *args, **kwargs)



from typing import List, Optional
import threading
from uuid import uuid4, UUID
from datetime import datetime, timezone
from django.utils import timezone as dj_timezone

from ..conf import settings
from ..const import TRACKING, LogOperate
from ..signals import log_created


__all__ = 'log', 'commit_log', 'create_log',


def log(
    domain: str,
    message: str = '',
    tracking: Optional[List[str]] = None,
    detail: Optional[dict] = None,
    operate: Optional[LogOperate] = None,
) -> UUID:
    id = uuid4()
    detail = {} if detail is None else detail
    tracking = TRACKING.get() if tracking is None else tracking
    created_at = datetime.utcnow().replace(tzinfo=timezone.utc)

    commit_log(
        operate=settings.OPERATE if operate is None else operate,
        kwargs={
            'id': id,
            'domain': domain,
            'tracking': tracking,
            'message': message,
            'detail': detail,
            'created_at': created_at
        }
    )

    return id


def commit_log(
    operate, args: list = [], kwargs: dict = {}
):
    if LogOperate.STRAIGHT == operate:
        return create_log(*args, **kwargs)

    if LogOperate.THREADING == operate:
        thread = threading.Thread(
            target=create_log,
            name='Logger',
            args=args,
            kwargs=kwargs,
            daemon=True
        )
        thread.start()


def create_log(id, domain, tracking, message, detail, created_at):
    instance = settings.get_saver()(
        id=id,
        domain=domain,
        tracking=tracking,
        message=message,
        detail=detail,
        created_at=created_at
    )
    log_created.send(settings.SAVER, instance=instance)

    return instance



from typing import List, Optional
from uuid import UUID
from px_domains import Domain
import datetime

from .models import Log
from .signals import log_stored


__all__ = 'save',


def save(
    id: UUID,
    domain: Domain,
    created_at: datetime.datetime,
    tracking: Optional[List[str]] = [],
    message: Optional[str] = '',
    detail: Optional[str] = '',
) -> Log:
    instance = Log.objects.create(
        id=id,
        domain=domain,
        tracking=tracking,
        message=message,
        detail=detail,
        created_at=created_at
    )

    log_stored.send(Log, instance=instance)

    return instance



from contextvars import ContextVar
from enum import Enum
from typing import List


__all__ = 'TRACKING', 'LogOperate'

TRACKING: ContextVar[List[str]] = ContextVar('logger.tracking', default=[])


class LogOperate(str, Enum):
    STRAIGHT = 'straight'
    THREADING = 'threading'



from django.contrib import admin
from django.utils.translation import pgettext_lazy
from django.db import models
from prettyjson import PrettyJSONWidget
from django.utils import timezone, formats

from .models import Log


@admin.register(Log)
class LogAdmin(admin.ModelAdmin):
    list_display = 'id', 'domain', 'message', 'tracking', 'created_at_extended',
    list_display_links = 'id', 'message',
    list_filter = (
        # 'domain',
        'created_at',
    )
    search_fields = 'id', 'domain', 'tracking', 'message', 'detail',
    readonly_fields = 'id', 'domain', 'tracking', 'message', 'created_at_extended',
    date_hierarchy = 'created_at'

    fieldsets = (
        (
            None,
            {
                'fields': (
                    ('domain', 'message'),
                    'detail',
                    'tracking',
                    'created_at_extended',
                )
            }
        ),
    )

    formfield_overrides = {models.JSONField: {'widget': PrettyJSONWidget}}

    def created_at_extended(self, obj):
        return '{} {}'.format(
            formats.localize(timezone.template_localtime(obj.created_at)),
            obj.created_at.strftime('%S.%f'),
        )

    created_at_extended.admin_order_field = 'created_at'
    created_at_extended.short_description = pgettext_lazy(
        'pxd-actions-tracker', 'Created at'
    )



from typing import Callable, Dict, Optional, Sequence, Tuple
import traceback
import logging
from uuid import UUID
from requests import Response, PreparedRequest, ConnectionError

from ..cases import log
from ..utils import clean_data
from ..domains import ERROR, INFO


__all__ = (
    'MESSAGE_FORMAT',
    'track',

    'should_track_non_200',
    'should_track_all',
    'resolve_log_detail',
)


logger = logging.getLogger(__name__)


def MESSAGE_FORMAT(**detail):
    return f'[REQUESTS]: "{detail.get("request", {}).get("url", "")}"'


def resolve_request_data(request: PreparedRequest) -> Dict:
    return {
        'method': request.method, 'url': request.url,
        'body': clean_data(request.body),
        'headers': clean_data(dict(request.headers)),
    }


def resolve_response_data(response: Response) -> Dict:
    return {
        'status_code': response.status_code,
        'content': clean_data(response.content),
    }


def resolve_exception_data(
    error: Optional[Exception],
    response: Optional[Response]
) -> Optional[Dict]:
    if error is not None:
        return {'message': str(error), 'traceback': traceback.format_exc()}

    return None


def should_track_non_200(
    request: Optional[PreparedRequest] = None,
    response: Optional[Response] = None,
    error: Optional[Exception] = None,
    args: Sequence = (),
    kwargs: Dict = {},
) -> Tuple[bool, str]:
    should = (
        error is not None
        or
        response is None
        or
        response.status_code // 100 in (4, 5)
    )

    return should, ERROR


def should_track_all(
    request: Optional[PreparedRequest] = None,
    response: Optional[Response] = None,
    error: Optional[Exception] = None,
    args: Sequence = (),
    kwargs: Dict = {},
) -> Tuple[bool, str]:
    e, _ = should_track_non_200(response=response, error=error)

    return True, ERROR if e else INFO


def resolve_log_detail(
    request: Optional[PreparedRequest] = None,
    response: Optional[Response] = None,
    error: Optional[Exception] = None,
    args: Sequence = (),
    kwargs: Dict = {},
) -> Dict:
    request = request if request else response.request if response is not None else None
    request_data = {} if request is None else resolve_request_data(request)
    response_data = None if response is None else resolve_response_data(response)
    error_data = resolve_exception_data(error, response)

    return {
        'request': request_data, 'response': response_data,
        'error': error_data,
    }


def track(
    domain: str,
    resolver: Callable,
    args: Sequence = (),
    kwargs: Dict = {},
    message_format: Callable = MESSAGE_FORMAT,
    should_track: Callable = should_track_non_200,
    detail_resolver: Callable = resolve_log_detail,
) -> Tuple[Optional[Response], Optional[UUID]]:
    error = None
    response: Optional[Response] = None
    request: Optional[PreparedRequest] = None

    try:
        response = resolver()

        if response is not None:
            request = response.request
    except ConnectionError as e:
        response = e.response
        request = e.request
        error = e
    except Exception as e:
        error = e

    kw = dict(
        request=request, response=response, error=error,
        args=args, kwargs=kwargs,
    )
    should, domain_postfix = should_track(**kw)

    if not should:
        return response, None

    id = None
    try:
        detail = detail_resolver(**kw)
        id = log(
            (domain | domain_postfix) if domain_postfix else domain,
            message_format(**detail), detail=detail,
        )
    except Exception as e:
        logger.exception(e)

    if error is not None:
        raise error

    return response, id



import ast
import ipaddress
from typing import Any, Optional, TypeVar

from .conf import settings


__all__ = (
    'is_list',
    'is_dict',
    'clean_data',
    'get_ip_address',
)

T = TypeVar('T')


def is_list(value) -> bool:
    return isinstance(value, (list, tuple))


def is_dict(value) -> bool:
    return isinstance(value, dict)


def clean_data(
    data: T,
    sensitive_fields: Optional[set] = None,
    sensitive_placeholder: Optional[str] = None,
) -> T:
    """Replaces sensitive information from data."""
    sensitive_fields = (
        settings.SENSITIVE_FIELDS
        if sensitive_fields is None else
        sensitive_fields
    )
    sensitive_placeholder = (
        settings.SENSITIVE_PLACEHOLDER
        if sensitive_placeholder is None else
        sensitive_placeholder
    )

    if isinstance(data, bytes):
        data = data.decode(errors='replace')

    if is_list(data):
        return [clean_data(d, sensitive_fields=sensitive_fields) for d in data]

    if is_dict(data):
        data = dict(data)

        for key, value in data.items():
            try:
                value = ast.literal_eval(value)
            except (ValueError, SyntaxError):
                pass

            if is_list(value) or is_dict(value):
                data[key] = clean_data(
                    value, sensitive_fields=sensitive_fields
                )

            if key.lower() in sensitive_fields:
                data[key] = sensitive_placeholder

    return data


def get_ip_address(request):
    """Get the remote ip address the request was generated from. """
    address = request.META.get('HTTP_X_FORWARDED_FOR', None)

    if address:
        address = address.split(',')[0]
    else:
        address = request.META.get('REMOTE_ADDR', '')

    # Account for IPv4 and IPv6 addresses, each possibly with port appended.
    # Possibilities are:
    # <ipv4 address>
    # <ipv6 address>
    # <ipv4 address>:port
    # [<ipv6 address>]:port
    # Note that ipv6 addresses are colon separated hex numbers
    possibles = (address.lstrip('[').split(']')[0], address.split(':')[0])

    for addr in possibles:
        try:
            return str(ipaddress.ip_address(addr))
        except ValueError:
            pass

    return address



import os
import re
import setuptools


with open('README.md', 'r') as rf:
    with open('CHANGELOG.md', 'r') as cf:
        long_description = rf.read() + cf.read()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()

    return re.search('__version__ = [\'"]([^\'"]+)[\'"]', init_py).group(1)


version = get_version('pxd_actions_tracker')


setuptools.setup(
    name='px-django-actions-tracker',
    version=version,
    author='Alex Tkachenko',
    author_email='preusx.dev@gmail.com',
    license='MIT License',
    description='Simple in-app actions tracker.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(exclude=(
        'tests', 'tests.*',
        'experiments', 'pilot',
    )),
    install_requires=(
        'Django',
        'px-settings',
        'px-domains',
        'django-prettyjson>=0.4.1,<0.5.0',
    ),
    extras_require={
        'dev': (
            'pytest>=6.0,<7.0',
            'pytest-watch>=4.2,<5.0',
            'pytest-django>=4.3,<5.0',
            'django-environ==0.4.5',
            'django-stubs',
            'django>=2.2,<4',
            'twine',
        ),
    },
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',

        'Programming Language :: Python :: 3',
        'Framework :: Django',

        'Intended Audience :: Developers',
        'Topic :: Utilities',

        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy

__all__ = ('ActionsTrackerConfig',)


class ActionsTrackerConfig(AppConfig):
    name = 'pxd_actions_tracker'
    verbose_name = pgettext_lazy('pxd-actions-tracker', 'Actions tracker')
